from threading import Thread

import testmod

class GPUCall(Thread):
    """

    """

    def __init__(self, func, device=0):
        self.func = func
        self.device = device

    def acquire_device(self):
        testmod.set_device(self.device)

    def release_device(self):
        pass

    def run(self):
        self.acquire_device()
        self.func()
        self.release_device()

def make_calls(func, data, devices=None, splits=None):
    """

    Parameters
    ----------

    Returns
    -------

    """
    if splits is None:
        pass

def _execute_calls(calls):
    """

    """
    for call in calls:
        call.start()

    for call in calls:
        call.join()







import nose
import sys
import unittest

from numpy.random import randn
from numpy.linalg import inv, cholesky as chol
from numpy.testing import assert_almost_equal, assert_equal
import numpy as np

import scipy.stats as sp_stats

import gpustats as gps
import gpustats.compat as compat
import gpustats.util as util

DECIMAL_6 = 6
DECIMAL_5 = 5
DECIMAL_4 = 4
DECIMAL_3 = 3
DECIMAL_2 = 2

def _make_test_case(n=1000, k=4, p=1):
    data = randn(n, k)
    covs = [util.random_cov(k) for _ in range(p)]
    means = [randn(k) for _ in range(p)]
    return data, means, covs

# debugging...

def _compare_multi(n, k, p):
    data, means, covs = _make_test_case(n, k, p)

    # cpu in PyMC
    pyresult = compat.python_mvnpdf(data, means, covs)

    # gpu
    result = gps.mvnpdf_multi(data, means, covs)

    return result, pyresult

def _compare_single(n, k):
    data, means, covs = _make_test_case(n, k, 1)

    mean = means[0]
    cov = covs[0]

    # cpu in PyMC
    pyresult = compat.python_mvnpdf(data, [mean], [cov]).squeeze()
    # gpu
    result = gps.mvnpdf(data, mean, cov)
    return result, pyresult

class TestMVN(unittest.TestCase):
    # ndata, dim, ncomponents
    test_cases = [(1000, 4, 1),
                  (1000, 4, 16),
                  (1000, 4, 32),
                  (1000, 4, 64),
                  (1000, 7, 64),
                  (1000, 8, 64),
                  (1000, 14, 32),
                  (1000, 16, 32),
                  (250, 25, 32),
                  (10, 15, 2)]

    def _check_multi(self, n, k, p):
        a, b = _compare_multi(n, k, p)
        assert_almost_equal(a, b, DECIMAL_2)

    def _check_single(self, n, k):
        a, b = _compare_single(n, k)
        assert_almost_equal(a, b, DECIMAL_2)

    def test_multi(self):
        for n, k, p in self.test_cases:
            self._check_multi(n, k, p)

    def test_single(self):
        for n, k, p in self.test_cases:
            self._check_single(n, k)

class TestUnivariate(unittest.TestCase):

    def test_normal(self):
        test_cases = [
            (100, 0, 1),
            (100, .5, 2.5),
            (10, 5, 3),
            (2000, 1, 4)
        ]

        for n, mean, std in test_cases:
            data = randn(n)
            pyresult = sp_stats.norm.pdf(data, loc=mean, scale=std)

            result = gps.normpdf(data, mean, std, logged=True)
            assert_almost_equal(result, np.log(pyresult), DECIMAL_5)

if __name__ == '__main__':
    # nose.runmodule(argv=['', '--pdb', '-v', '--pdb-failure'])
    pass



from gpustats.codegen import MVDensityKernel, DensityKernel, Exp
import gpustats.codegen as cg

# TODO: check for name conflicts!

_log_pdf_mvnormal = """
__device__ float %(name)s(float* data, float* params, int dim) {
  unsigned int LOGDET_OFFSET = dim * (dim + 3) / 2;
  float* mean = params;
  float* sigma = params + dim;
  float mult = params[LOGDET_OFFSET];
  float logdet = params[LOGDET_OFFSET + 1];

  float discrim = 0;
  float sum;
  unsigned int i, j;
  for (i = 0; i < dim; ++i)
  {
    sum = 0;
    for(j = 0; j <= i; ++j) {
      sum += *sigma++ * (data[j] - mean[j]);
    }
    discrim += sum * sum;
  }
  return log(mult) - 0.5 * (discrim + logdet + LOG_2_PI * dim);
}
"""
log_pdf_mvnormal = MVDensityKernel('log_pdf_mvnormal', _log_pdf_mvnormal)
pdf_mvnormal = Exp('pdf_mvnormal', log_pdf_mvnormal)


_log_pdf_normal = """
__device__ float %(name)s(float* x, float* params) {
  // mean stored in params[0]
  float std = params[1];

  // standardize
  float xstd = (*x - params[0]) / std;
  return - (xstd * xstd) / 2 - 0.5 * LOG_2_PI - log(std);
}
"""
log_pdf_normal = DensityKernel('log_pdf_normal', _log_pdf_normal)
pdf_normal = Exp('pdf_normal', log_pdf_normal)



"""
Python versions of functions for testing purposes etc.
"""
import numpy as np
import pymc.distributions as pymc_dist

def python_mvnpdf(data, means, covs):
    pdf_func = pymc_dist.mv_normal_cov_like

    results = []
    for i, datum in enumerate(data):
        for j, cov in enumerate(covs):
            mean = means[j]
            results.append(pdf_func(datum, mean, cov))

    return np.array(results).reshape((len(data), len(covs))).squeeze()



from numpy.random import randn
from numpy.linalg import cholesky as chol
import numpy as np
import numpy.linalg as LA

import gpustats.kernels as kernels
import gpustats.codegen as codegen
reload(codegen)
reload(kernels)
import gpustats.util as util
import pycuda.driver as drv

__all__ = ['mvnpdf', 'mvnpdf_multi', 'normpdf', 'normpdf_multi']

cu_module = codegen.get_full_cuda_module()

#-------------------------------------------------------------------------------
# Invokers for univariate and multivariate density functions conforming to the
# standard API

def _multivariate_pdf_call(cu_func, data, packed_params):
    padded_data = util.pad_data(data)
    packed_params = util.prep_ndarray(packed_params)

    ndata, dim = data.shape

    nparams = len(packed_params)
    data_per, params_per = util.tune_blocksize(padded_data,
                                               packed_params)

    shared_mem = util.compute_shmem(padded_data, packed_params,
                                    data_per, params_per)

    block_design = (data_per * params_per, 1, 1)
    grid_design = (util.get_boxes(ndata, data_per),
                   util.get_boxes(nparams, params_per))

    # see cufiles/mvcaller.cu
    design = np.array(((data_per, params_per) + # block design
                       padded_data.shape + # data spec
                       (dim,) + # non-padded number of data columns
                       packed_params.shape), # params spec
                      dtype=np.float32)

    if nparams == 1:
        dest = np.zeros(ndata, dtype=np.float32)
    else:
        dest = np.zeros((ndata, nparams), dtype=np.float32, order='F')

    cu_func(drv.Out(dest),
            drv.In(padded_data), drv.In(packed_params), drv.In(design),
            block=block_design, grid=grid_design, shared=shared_mem)

    return dest

def _univariate_pdf_call(cu_func, data, packed_params):
    ndata = len(data)
    nparams = len(packed_params)

    data = util.prep_ndarray(data)
    packed_params = util.prep_ndarray(packed_params)

    data_per, params_per = util.tune_blocksize(data, packed_params)

    shared_mem = util.compute_shmem(data, packed_params,
                                    data_per, params_per)

    block_design = (data_per * params_per, 1, 1)
    grid_design = (util.get_boxes(ndata, data_per),
                   util.get_boxes(nparams, params_per))

    # see cufiles/univcaller.cu
    design = np.array(((data_per, params_per) + # block design
                       (len(data),) +
                       packed_params.shape), # params spec
                      dtype=np.float32)

    dest = np.zeros((ndata, nparams), dtype=np.float32, order='F')

    cu_func(drv.Out(dest),
            drv.In(data), drv.In(packed_params), drv.In(design),
            block=block_design, grid=grid_design, shared=shared_mem)

    return dest

#-------------------------------------------------------------------------------
# Multivariate normal

def mvnpdf(data, mean, cov, weight=None, logged=True):
    """
    Multivariate normal density

    Parameters
    ----------

    Returns
    -------
    """
    return mvnpdf_multi(data, [mean], [cov]).squeeze()

def mvnpdf_multi(data, means, covs, weights=None, logged=True):
    """
    Multivariate normal density with multiple sets of parameters

    Parameters
    ----------
    data : ndarray (n x k)
    covs : sequence of 2d k x k matrices (length j)
    weights : ndarray (length j)
        Multiplier for component j, usually will sum to 1

    Returns
    -------
    densities : n x j
    """
    if logged:
        cu_func = cu_module.get_function('log_pdf_mvnormal')
    else:
        cu_func = cu_module.get_function('pdf_mvnormal')

    ichol_sigmas = [LA.inv(chol(c)) for c in covs]
    logdets = [np.log(LA.det(c)) for c in covs]

    packed_params = _pack_mvnpdf_params(means, ichol_sigmas, logdets)

    return _multivariate_pdf_call(cu_func, data, packed_params)

def _pack_mvnpdf_params(means, ichol_sigmas, logdets):
    to_pack = []
    for m, ch, ld in zip(means, ichol_sigmas, logdets):
        to_pack.append(_pack_mvnpdf_params_single(m, ch, ld))

    return np.vstack(to_pack)

def _pack_mvnpdf_params_single(mean, ichol_sigma, logdet):
    PAD_MULTIPLE = 16
    k = len(mean)
    mean_len = k
    ichol_len = k * (k + 1) / 2
    mch_len = mean_len + ichol_len

    packed_dim = util.next_multiple(mch_len + 2, PAD_MULTIPLE)

    packed_params = np.empty(packed_dim, dtype=np.float32)
    packed_params[:mean_len] = mean

    packed_params[mean_len:mch_len] = ichol_sigma[np.tril_indices(k)]
    packed_params[mch_len:mch_len + 2] = 1, logdet

    return packed_params

#-------------------------------------------------------------------------------
# Univariate normal

def normpdf(x, mean, std, logged=True):
    """
    Normal (Gaussian) density

    Parameters
    ----------

    Returns
    -------
    """
    return normpdf_multi(x, [mean], [std], logged=logged).squeeze()

def normpdf_multi(x, means, std, logged=True):
    if logged:
        cu_func = cu_module.get_function('log_pdf_normal')
    else:
        cu_func = cu_module.get_function('pdf_normal')

    packed_params = np.c_[means, std]
    return _univariate_pdf_call(cu_func, x, packed_params)

if __name__ == '__main__':
    import gpustats.compat as compat

    n = 1e5
    k = 8

    np.random.seed(1)
    data = randn(n, k).astype(np.float32)
    mean = randn(k).astype(np.float32)
    cov = util.random_cov(k).astype(np.float32)

    result = mvnpdf(data, mean, cov)
    pyresult = compat.python_mvnpdf(data, [mean], [cov], 8).squeeze()

    print result - pyresult



import pycuda.driver as drv
import pycuda.tools
import pycuda.autoinit
import numpy
import numpy.linalg as la
import os
from pycuda.compiler import SourceModule

class CUDAModule(object):
    """
    Interfaces with PyCUDA

    Parameters
    ----------
    kernel_dict :
    """
    def __init__(self, kernel_dict):
        self.kernel_dict = kernel_dict
        self.support_code = _get_support_code()

        all_code = self._get_full_source()
        try:
            self.pycuda_module = SourceModule(all_code)
        except Exception:
            f = open('foo.cu', 'w')
            print >> f, all_code
            f.close()
            raise

    def _get_full_source(self):
        formatted_kernels = []

        for name, kernel in self.kernel_dict.iteritems():
            logic = kernel.get_logic()
            caller = kernel.get_caller()
            complete_code = '\n'.join((logic, caller))

            formatted_kernels.append(complete_code)

        return '\n'.join([self.support_code] + formatted_kernels)

    def get_function(self, name):
        return self.pycuda_module.get_function('k_%s' % name)

def _get_support_code():
    path = os.path.join(_get_cuda_code_path(), 'support.cu')
    return open(path).read()

def _get_mvcaller_code():
    # for multivariate pdfs
    path = os.path.join(_get_cuda_code_path(), 'mvcaller.cu')
    return open(path).read()

def _get_univcaller_code():
    # For univariate pdfs
    path = os.path.join(_get_cuda_code_path(), 'univcaller.cu')
    return open(path).read()

def _get_cuda_code_path():
    import os.path as pth
    basepath = pth.abspath(pth.split(__file__)[0])
    return pth.join(basepath, 'cufiles')

class Kernel(object):

    def __init__(self, name):
        if name is None:
            raise ValueError('Kernel must have a default name')

        self.name = name

    def get_name(self, name=None):
        # can override default name, for transforms. this a hack?
        if name is None:
            name = self.name

        return name

class DensityKernel(Kernel):
    """
    Generate kernel for probability density function
    """

    _caller = _get_univcaller_code()
    def __init__(self, name, logic_code):

        self.logic_code = logic_code

        Kernel.__init__(self, name)

    def get_logic(self, name=None):
        return self.logic_code % {'name' : self.get_name(name)}

    def get_caller(self, name=None):
        return self._caller % {'name' : self.get_name(name)}

class MVDensityKernel(DensityKernel):
    """

    """
    _caller = _get_mvcaller_code()

class Transform(Kernel):
    """
    Enable simple transforms of kernels to compute modified kernel code stub
    """
    def __init__(self, name, kernel):
        self.kernel = kernel
        Kernel.__init__(self, name)

    # XXX: HACK, not general for non-density kernels
    def is_multivariate(self):
        return isinstance(self.kernel, MVDensityKernel)

# flop the right name?

class Flop(Transform):
    op = None

    def get_logic(self, name=None):
        name = self.get_name(name)

        actual_name = '%s_stub' % name
        kernel_logic = self.kernel.get_logic(name=actual_name)

        if self.is_multivariate():
            stub_caller = _mv_stub_caller
        else:
            stub_caller = _univ_stub_caller

        transform_logic = stub_caller % {'name' : name,
                                         'actual_kernel' : actual_name,
                                         'op' : self.op}

        return '\n'.join((kernel_logic, transform_logic))

    def get_caller(self):
        return self.kernel.get_caller(self.name)

_univ_stub_caller = """
__device__ float %(name)s(float* x, float* params) {
    return %(op)s(%(actual_kernel)s(x, params));
}
"""

_mv_stub_caller = """
__device__ float %(name)s(float* x, float* params, int dim) {
    return %(op)s(%(actual_kernel)s(x, params, dim));
}
"""

class Exp(Flop):
    op = 'exp'

class Log(Flop):
    op = 'log'

class Sqrt(Flop):
    op = 'sqrt'

def get_full_cuda_module():
    import gpustats.kernels as kernels
    objects = kernels.__dict__

    all_kernels = dict((k, v)
                       for k, v in kernels.__dict__.iteritems()
                       if isinstance(v, Kernel))
    return CUDAModule(all_kernels)

if __name__ == '__main__':
    pass



#/usr/bin/env python

from numpy.distutils.misc_util import Configuration
from numpy.distutils.core import setup

DESCRIPTION = "GPU-based statistical functions"
LONG_DESCRIPTION = """
gpustats is a PyCUDA-based library implementing functionality similar to that
present in scipy.stats. It implements a simple framework for specifying new CUDA
kernels and extending existing ones. Here is a (partial) list of target
functionality:

* Probability density functions (pdfs). These are intended to speed up
  likelihood calculations in particular in Bayesian inference applications, such
  as in PyMC

* Random variable generation using CURAND

Notes
-----
Requires working PyCUDA installation
"""

REQUIRES = ['numpy', 'pycuda >= 0.94rc']
DISTNAME = 'gpustats'
LICENSE = 'BSD'
AUTHOR = "Wes McKinney"
AUTHOR_EMAIL = "wesmckinn@gmail.com"
URL = "https://github.com/dukestats/gpustats"
CLASSIFIERS = [
    'Development Status :: 2 - Pre-Alpha',
    'Environment :: Console',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
]

MAJOR = 0
MINOR = 0
MICRO = 1
ISRELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

FULLVERSION = VERSION
if not ISRELEASED:
    FULLVERSION += '.beta'

def configuration(parent_package='', top_path=None):
    config = Configuration(None, parent_package, top_path,
                           version=FULLVERSION)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)

    config.add_subpackage('gpustats')
    config.add_data_dir('gpustats/tests')
    config.add_data_dir('gpustats/cufiles')
    return config

if __name__ == '__main__':
    setup(name=DISTNAME,
          author=AUTHOR,
          author_email=AUTHOR_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          long_description=LONG_DESCRIPTION,
          classifiers=CLASSIFIERS,
          platforms='any',
          configuration=configuration)



import numpy as np
import pymc.distributions as pymc_dist
import pycuda.driver as drv

_dev_attr = drv.device_attribute

class DeviceInfo(object):

    def __init__(self, dev=0):
        self.dev = dev
        self._dev = drv.Device(dev)
        self._attr = self._dev.get_attributes()

        self.max_block_threads = self._attr[_dev_attr.MAX_THREADS_PER_BLOCK]
        self.shared_mem = self._attr[_dev_attr.MAX_SHARED_MEMORY_PER_BLOCK]
        self.warp_size = self._attr[_dev_attr.WARP_SIZE]

HALF_WARP = 16

def random_cov(dim):
    return pymc_dist.rinverse_wishart(dim, np.eye(dim))

def unvech(v):
    # quadratic formula, correct fp error
    rows = .5 * (-1 + np.sqrt(1 + 8 * len(v)))
    rows = int(np.round(rows))

    result = np.zeros((rows, rows))
    result[np.triu_indices(rows)] = v
    result = result + result.T

    # divide diagonal elements by 2
    result[np.diag_indices(rows)] /= 2

    return result

def pad_data(data):
    """
    Pad data to avoid bank conflicts on the GPU-- dimension should not be a
    multiple of the half-warp size (16)
    """
    n, k = data.shape

    if not k % HALF_WARP:
        pad_dim = k + 1
    else:
        pad_dim = k

    if k != pad_dim:
        padded_data = np.empty((n, pad_dim), dtype=np.float32)
        padded_data[:, :k] = data

        return padded_data
    else:
        return prep_ndarray(data)

def prep_ndarray(arr):
    # is float32 and contiguous?
    if not arr.dtype == np.float32 or not arr.flags.contiguous:
        arr = np.array(arr, dtype=np.float32)

    return arr


def tune_blocksize(data, params, device=0):
    """
    For multivariate distributions-- what's the optimal block size given the
    gpu?

    Parameters
    ----------
    data : ndarray
    params : ndarray

    Returns
    -------
    (data_per, params_per) : (int, int)
    """
    # TODO: how to figure out active device in this thread for the multigpu
    # case?
    info = DeviceInfo(device)

    max_smem = info.shared_mem * 0.9
    max_threads = info.max_block_threads

    params_per = max_threads
    if (len(params) < params_per):
        params_per = _next_pow2(len(params), info.max_block_threads)

    data_per = max_threads / params_per

    def _can_fit(data_per, params_per):
        return compute_shmem(data, params, data_per, params_per) <= max_smem

    while True:
        while not _can_fit(data_per, params_per):
            if data_per <= 1:
                break

            if params_per > 1:
                # reduce number of parameters first
                params_per /= 2
            else:
                # can't go any further, have to do less data
                data_per /= 2

        if data_per == 0:
            # we failed somehow. start over
            data_per = 1
            params_per /= 2
            continue
        else:
            break

    while _can_fit(2 * data_per, params_per):
        if 2 * data_per * params_per < max_threads:
            data_per *= 2
        else:
            # hit block size limit
            break

    return data_per, params_per

def get_boxes(n, box_size):
    # how many boxes of size box_size are needed to hold n things
    return int((n + box_size - 1) / box_size)

def compute_shmem(data, params, data_per, params_per):
    result_space = data_per * params_per

    data_dim = 1 if data.ndim == 1 else data.shape[1]
    params_dim = len(params) if params.ndim == 1 else params.shape[1]

    param_space = params_dim * params_per
    data_space = data_dim * data_per
    return 4 * (result_space + param_space + data_space)

def _next_pow2(k, pow2):
    while k <= pow2 / 2:
        pow2 /= 2
    return pow2

def next_multiple(k, mult):
    if k % mult:
        return k + (mult - k % mult)
    else:
        return k
